/*
 * Copyright (C) 2016 Guido Berhoerster <guido+pwm@berhoerster.name>
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/types.h>
#include <errno.h>

#include "getline.h"

#define	MIN(x, y)	(((x) < (y)) ? (x) : (y))

ssize_t
getdelim(char **linep, size_t *sizep, int delim, FILE *fp)
{
	ssize_t	len;
	char	*line = *linep;
	size_t	size = *sizep;
	size_t	pos = 0;
	char	*line_new;
	int	c;

	if (*linep == NULL) {
		size = 0;
	}

	flockfile(fp);

	for (;;) {
		if (size - pos < 2) {
			size = MIN((size_t)SSIZE_MAX + 1,
			    (size < BUFSIZ) ? BUFSIZ : size * 1.5);
			if (size - pos < 2) {
				len = -1;
				errno = EOVERFLOW;
				break;
			}
			line_new = realloc(line, size);
			if (line_new == NULL) {
				len = -1;
				break;
			}
			*linep = line = line_new;
			*sizep = size;
		}
		c = fgetc(fp);
		if (c == EOF) {
			len = -1;
			break;
		}
		line[pos] = c;
		pos++;
		len = pos;
		if (c == delim) {
			break;
		}
	}

	if (*linep != NULL) {
		(*linep)[pos] = '\0';
	}

	funlockfile(fp);

	return (len);
}

ssize_t
getline(char **linep, size_t *sizep, FILE *fp)
{
	return (getdelim(linep, sizep, '\n', fp));
}
