/*
 * Copyright (C) 2017 Guido Berhoerster <guido+pwm@berhoerster.name>
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "compat.h"

#ifdef	HAVE_ERR_H
#include <err.h>
#endif /* HAVE_ERR_H */
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>

#include "rand.h"

#ifdef	__linux__
#define	PATH_DEV_RANDOM	"/dev/urandom"
#else
#define	PATH_DEV_RANDOM	"/dev/random"
#endif /* __linux__ */

void
rand_buf(void *buf, size_t buf_size)
{
	unsigned char	*p = buf;
	int		fd;
	ssize_t		nread;
	size_t		nsize = buf_size;

	do {
		fd = open(PATH_DEV_RANDOM, O_RDONLY);
	} while ((fd < 0) && (errno == EINTR));
	if (fd < 0) {
		err(1, "open");
	}

	while (nsize > 0) {
		nread = read(fd, p, nsize);
		if (nread < 0) {
			if (errno == EINTR) {
				continue;
			}
			err(1, "read");
		}
		p += nread;
		nsize -= nread;
	}

	while ((close(fd) < 0) && (errno == EINTR));
}

uint32_t
rand_random(void)
{
	uint32_t	x;

	rand_buf(&x, sizeof (x));

	return (x);
}

/* random number between 0 and upper_bound - 1 without modulo bias */
uint32_t
rand_uniform(uint32_t upper_bound)
{
	uint32_t	r;
	/* (2^32 - upper_bound) % upper_bound */
	uint32_t	threshold = -upper_bound % upper_bound;

	do {
		r = rand_random();
	} while (r < threshold);

	return (r % upper_bound);
}
